@@ -21,6 +21,8 @@
 #endif
 
 
+static const char *basepath;
+static const char *basepath_r;
 static char testfile[1024];
 static char testfile2[1024];
 static char testdir[1024];
@@ -42,11 +44,18 @@ static long seekdir_offsets[4];
 static char zerodata[4096];
 static int testdatalen = sizeof(testdata) - 1;
 static int testdata2len = sizeof(testdata2) - 1;
-static unsigned int testnum = 1;
+static unsigned int testnum = 0;
 static unsigned int select_test = 0;
 static unsigned int skip_test = 0;
+static unsigned int unlinked_test = 0;
 
 #define MAX_ENTRIES 1024
+#define MAX_TESTS 100
+
+static struct test {
+	int fd;
+	struct stat stat;
+} tests[MAX_TESTS];
 
 static void test_perror(const char *func, const char *msg)
 {
@@ -80,21 +89,32 @@ static void success(void)
 	fprintf(stderr, "%s OK\n", testname);
 }
 
+#define this_test (&tests[testnum-1])
+#define next_test (&tests[testnum])
+
 static void __start_test(const char *fmt, ...)
 {
 	unsigned int n;
 	va_list ap;
-	n = sprintf(testname, "%3i [", testnum++);
+	n = sprintf(testname, "%3i [", testnum);
 	va_start(ap, fmt);
 	n += vsprintf(testname + n, fmt, ap);
 	va_end(ap);
 	sprintf(testname + n, "]");
+	// Use dedicated testfile per test
+	sprintf(testfile, "%s/testfile.%d", basepath, testnum);
+	sprintf(testfile_r, "%s/testfile.%d", basepath_r, testnum);
+	if (testnum > MAX_TESTS) {
+		fprintf(stderr, "%s - too many tests\n", testname);
+		exit(1);
+	}
+	this_test->fd = -1;
 }
 
 #define start_test(msg, args...) { \
+	testnum++; \
 	if ((select_test && testnum != select_test) || \
 	    (testnum == skip_test)) { \
-		testnum++; \
 		return 0; \
 	} \
 	__start_test(msg, ##args);		\
@@ -105,6 +125,16 @@ static void __start_test(const char *fmt, ...)
 
 #define ARRAY_SIZE(arr) (sizeof(arr) / sizeof((arr)[0]))
 
+static int st_check_size(struct stat *st, int len)
+{
+	if (st->st_size != len) {
+		ERROR("length %u instead of %u", (int) st->st_size,
+		      (int) len);
+		return -1;
+	}
+	return 0;
+}
+
 static int check_size(const char *path, int len)
 {
 	struct stat stbuf;
@@ -113,25 +143,19 @@ static int check_size(const char *path, int len)
 		PERROR("stat");
 		return -1;
 	}
-	if (stbuf.st_size != len) {
-		ERROR("length %u instead of %u", (int) stbuf.st_size,
-		      (int) len);
-		return -1;
-	}
-	return 0;
+	return st_check_size(&stbuf, len);
 }
 
-static int fcheck_size(int fd, int len)
+static int check_testfile_size(const char *path, int len)
 {
-	struct stat stbuf;
-	int res = fstat(fd, &stbuf);
-	if (res == -1) {
-		PERROR("fstat");
-		return -1;
-	}
-	if (stbuf.st_size != len) {
-		ERROR("length %u instead of %u", (int) stbuf.st_size,
-		      (int) len);
+	this_test->stat.st_size = len;
+	return check_size(path, len);
+}
+
+static int st_check_type(struct stat *st, mode_t type)
+{
+	if ((st->st_mode & S_IFMT) != type) {
+		ERROR("type 0%o instead of 0%o", st->st_mode & S_IFMT, type);
 		return -1;
 	}
 	return 0;
@@ -145,23 +169,14 @@ static int check_type(const char *path, mode_t type)
 		PERROR("lstat");
 		return -1;
 	}
-	if ((stbuf.st_mode & S_IFMT) != type) {
-		ERROR("type 0%o instead of 0%o", stbuf.st_mode & S_IFMT, type);
-		return -1;
-	}
-	return 0;
+	return st_check_type(&stbuf, type);
 }
 
-static int fcheck_type(int fd, mode_t type)
+static int st_check_mode(struct stat *st, mode_t mode)
 {
-	struct stat stbuf;
-	int res = fstat(fd, &stbuf);
-	if (res == -1) {
-		PERROR("fstat");
-		return -1;
-	}
-	if ((stbuf.st_mode & S_IFMT) != type) {
-		ERROR("type 0%o instead of 0%o", stbuf.st_mode & S_IFMT, type);
+	if ((st->st_mode & ALLPERMS) != mode) {
+		ERROR("mode 0%o instead of 0%o", st->st_mode & ALLPERMS,
+		      mode);
 		return -1;
 	}
 	return 0;
@@ -175,28 +190,14 @@ static int check_mode(const char *path, mode_t mode)
 		PERROR("lstat");
 		return -1;
 	}
-	if ((stbuf.st_mode & ALLPERMS) != mode) {
-		ERROR("mode 0%o instead of 0%o", stbuf.st_mode & ALLPERMS,
-		      mode);
-		return -1;
-	}
-	return 0;
+	return st_check_mode(&stbuf, mode);
 }
 
-static int fcheck_mode(int fd, mode_t mode)
+static int check_testfile_mode(const char *path, mode_t mode)
 {
-	struct stat stbuf;
-	int res = fstat(fd, &stbuf);
-	if (res == -1) {
-		PERROR("fstat");
-		return -1;
-	}
-	if ((stbuf.st_mode & ALLPERMS) != mode) {
-		ERROR("mode 0%o instead of 0%o", stbuf.st_mode & ALLPERMS,
-		      mode);
-		return -1;
-	}
-	return 0;
+	this_test->stat.st_mode &= ~ALLPERMS;
+	this_test->stat.st_mode |= mode;
+	return check_mode(path, mode);
 }
 
 static int check_times(const char *path, time_t atime, time_t mtime)
@@ -247,6 +248,16 @@ static int fcheck_times(int fd, time_t atime, time_t mtime)
 }
 #endif
 
+static int st_check_nlink(struct stat *st, nlink_t nlink)
+{
+	if (st->st_nlink != nlink) {
+		ERROR("nlink %li instead of %li", (long) st->st_nlink,
+		      (long) nlink);
+		return -1;
+	}
+	return 0;
+}
+
 static int check_nlink(const char *path, nlink_t nlink)
 {
 	struct stat stbuf;
@@ -255,28 +266,31 @@ static int check_nlink(const char *path, nlink_t nlink)
 		PERROR("lstat");
 		return -1;
 	}
-	if (stbuf.st_nlink != nlink) {
-		ERROR("nlink %li instead of %li", (long) stbuf.st_nlink,
-		      (long) nlink);
-		return -1;
-	}
-	return 0;
+	return st_check_nlink(&stbuf, nlink);
 }
 
-static int fcheck_nlink(int fd, nlink_t nlink)
+static int fcheck_stat(int fd, int flags, struct stat *st)
 {
 	struct stat stbuf;
 	int res = fstat(fd, &stbuf);
 	if (res == -1) {
+		if (flags & O_PATH) {
+			// With O_PATH fd, the server does not have to keep
+			// the inode alive so FUSE inode may be stale or bad
+			if (errno == ESTALE || errno == EIO || errno == ENOENT)
+				return 0;
+		}
 		PERROR("fstat");
 		return -1;
 	}
-	if (stbuf.st_nlink != nlink) {
-		ERROR("nlink %li instead of %li", (long) stbuf.st_nlink,
-		      (long) nlink);
-		return -1;
-	}
-	return 0;
+
+	int err = 0;
+	err += st_check_type(&stbuf, st->st_mode & S_IFMT);
+	err += st_check_mode(&stbuf, st->st_mode & ALLPERMS);
+	err += st_check_size(&stbuf, st->st_size);
+	err += st_check_nlink(&stbuf, st->st_nlink);
+
+	return err;
 }
 
 static int check_nonexist(const char *path)
@@ -496,6 +510,90 @@ static int create_file(const char *path, const char *data, int len)
 	return 0;
 }
 
+static int create_path_fd(const char *path, const char *data, int len)
+{
+	int path_fd;
+	int res;
+
+	res = create_file(path, data, len);
+	if (res == -1)
+		return -1;
+
+	path_fd = open(path, O_PATH);
+	if (path_fd == -1)
+		PERROR("open(O_PATH)");
+
+	return path_fd;
+}
+
+// Can be called once per test
+static int create_testfile(const char *path, const char *data, int len)
+{
+	struct test *t = this_test;
+	struct stat *st = &t->stat;
+	int res, fd;
+
+	if (t->fd > 0) {
+		ERROR("testfile already created");
+		return -1;
+	}
+
+	fd = create_path_fd(path, data, len);
+	if (fd == -1)
+		return -1;
+
+	t->fd = fd;
+
+	res = fstat(fd, st);
+	if (res == -1) {
+		PERROR("fstat");
+		return -1;
+	}
+
+	return 0;
+}
+
+static int check_unlinked_testfile(int fd)
+{
+	struct stat *st = &this_test->stat;
+
+	st->st_nlink = 0;
+	return fcheck_stat(fd, O_PATH, st);
+}
+
+// Check recorded testfiles after all tests completed
+static int check_unlinked_testfiles(void)
+{
+	int fd;
+	int res, err = 0;
+	int num = testnum;
+
+	if (!unlinked_test)
+		return 0;
+
+	testnum = 0;
+	while (testnum < num) {
+		fd = next_test->fd;
+		start_test("check_unlinked_testfile");
+		if (fd == -1)
+			continue;
+
+		err += check_unlinked_testfile(fd);
+		res = close(fd);
+		if (res == -1) {
+			PERROR("close(test_fd)");
+			err--;
+		}
+	}
+
+	if (err) {
+		fprintf(stderr, "%i unlinked testfile checks failed\n", -err);
+		return 1;
+	}
+
+	return err;
+}
+
 static int cleanup_dir(const char *path, const char **dir_files, int quiet)
 {
 	int i;
@@ -560,7 +658,7 @@ static int test_truncate(int len)
 	int res;
 
 	start_test("truncate(%u)", (int) len);
-	res = create_file(testfile, data, datalen);
+	res = create_testfile(testfile, data, datalen);
 	if (res == -1)
 		return -1;
 
@@ -569,7 +667,7 @@ static int test_truncate(int len)
 		PERROR("truncate");
 		return -1;
 	}
-	res = check_size(testfile, len);
+	res = check_testfile_size(testfile, len);
 	if (res == -1)
 		return -1;
 
@@ -609,7 +707,7 @@ static int test_ftruncate(int len, int mode)
 	int fd;
 
 	start_test("ftruncate(%u) mode: 0%03o", len, mode);
-	res = create_file(testfile, data, datalen);
+	res = create_testfile(testfile, data, datalen);
 	if (res == -1)
 		return -1;
 
@@ -625,7 +723,7 @@ static int test_ftruncate(int len, int mode)
 		close(fd);
 		return -1;
 	}
-	res = check_mode(testfile, mode);
+	res = check_testfile_mode(testfile, mode);
 	if (res == -1) {
 		close(fd);
 		return -1;
@@ -637,7 +735,7 @@ static int test_ftruncate(int len, int mode)
 		return -1;
 	}
 	close(fd);
-	res = check_size(testfile, len);
+	res = check_testfile_size(testfile, len);
 	if (res == -1)
 		return -1;
 
@@ -830,7 +928,7 @@ static int test_utime(void)
 	int res;
 
 	start_test("utime");
-	res = create_file(testfile, NULL, 0);
+	res = create_testfile(testfile, NULL, 0);
 	if (res == -1)
 		return -1;
 
@@ -948,10 +1046,11 @@ static int test_create_unlink(void)
 		close(fd);
 		return -1;
 	}
-	err += fcheck_type(fd, S_IFREG);
-	err += fcheck_mode(fd, 0644);
-	err += fcheck_nlink(fd, 0);
-	err += fcheck_size(fd, datalen);
+	struct stat st = {
+		.st_mode = S_IFREG | 0644,
+		.st_size = datalen,
+	};
+	err = fcheck_stat(fd, O_RDWR, &st);
 	err += fcheck_data(fd, data, 0, datalen);
 	res = close(fd);
 	if (res == -1) {
@@ -1171,7 +1270,7 @@ static int do_test_open_acc(int flags, const char *flags_str, int mode, int err)
 	start_test("open_acc(%s) mode: 0%03o message: '%s'", flags_str, mode,
 		   strerror(err));
 	unlink(testfile);
-	res = create_file(testfile, data, datalen);
+	res = create_testfile(testfile, data, datalen);
 	if (res == -1)
 		return -1;
 
@@ -1181,7 +1280,7 @@ static int do_test_open_acc(int flags, const char *flags_str, int mode, int err)
 		return -1;
 	}
 
-	res = check_mode(testfile, mode);
+	res = check_testfile_mode(testfile, mode);
 	if (res == -1)
 		return -1;
 
@@ -1199,6 +1298,19 @@ static int do_test_open_acc(int flags, const char *flags_str, int mode, int err)
 		}
 		close(fd);
 	}
+
+	res = unlink(testfile);
+	if (res == -1) {
+		PERROR("unlink");
+		return -1;
+	}
+	res = check_nonexist(testfile);
+	if (res == -1)
+		return -1;
+	res = check_nonexist(testfile_r);
+	if (res == -1)
+		return -1;
+
 	success();
 	return 0;
 }
@@ -1213,7 +1325,7 @@ static int test_symlink(void)
 	int res;
 
 	start_test("symlink");
-	res = create_file(testfile, data, datalen);
+	res = create_testfile(testfile, data, datalen);
 	if (res == -1)
 		return -1;
 
@@ -1254,6 +1366,15 @@ static int test_symlink(void)
 	if (err)
 		return -1;
 
+	res = unlink(testfile);
+	if (res == -1) {
+		PERROR("unlink");
+		return -1;
+	}
+	res = check_nonexist(testfile);
+	if (res == -1)
+		return -1;
+
 	success();
 	return 0;
 }
@@ -1266,7 +1387,7 @@ static int test_link(void)
 	int res;
 
 	start_test("link");
-	res = create_file(testfile, data, datalen);
+	res = create_testfile(testfile, data, datalen);
 	if (res == -1)
 		return -1;
 
@@ -1316,7 +1437,7 @@ static int test_link2(void)
 	int res;
 
 	start_test("link-unlink-link");
-	res = create_file(testfile, data, datalen);
+	res = create_testfile(testfile, data, datalen);
 	if (res == -1)
 		return -1;
 
@@ -1375,7 +1496,7 @@ static int test_rename_file(void)
 	int res;
 
 	start_test("rename file");
-	res = create_file(testfile, data, datalen);
+	res = create_testfile(testfile, data, datalen);
 	if (res == -1)
 		return -1;
 
@@ -1835,39 +1956,42 @@ static int do_test_create_ro_dir(int flags, const char *flags_str)
 
 int main(int argc, char *argv[])
 {
-	const char *basepath;
-	const char *realpath;
 	int err = 0;
 	int a;
 	int is_root;
 
 	umask(0);
 	if (argc < 2 || argc > 4) {
-		fprintf(stderr, "usage: %s testdir [:realdir] [[-]test#]\n", argv[0]);
+		fprintf(stderr, "usage: %s testdir [:realdir] [[-]test#] [-u]\n", argv[0]);
 		return 1;
 	}
 	basepath = argv[1];
-	realpath = basepath;
+	basepath_r = basepath;
 	for (a = 2; a < argc; a++) {
 		char *endptr;
 		char *arg = argv[a];
 		if (arg[0] == ':') {
-			realpath = arg + 1;
+			basepath_r = arg + 1;
 		} else {
 			if (arg[0] == '-') {
 				arg++;
-				skip_test = strtoul(arg, &endptr, 10);
+				if (arg[0] == 'u') {
+					unlinked_test = 1;
+					endptr = arg + 1;
+				} else {
+					skip_test = strtoul(arg, &endptr, 10);
+				}
 			} else {
 				select_test = strtoul(arg, &endptr, 10);
 			}
 			if (arg[0] == '\0' || *endptr != '\0') {
-				fprintf(stderr, "invalid number: '%s'\n", arg);
+				fprintf(stderr, "invalid option: '%s'\n", argv[a]);
 				return 1;
 			}
 		}
 	}
 	assert(strlen(basepath) < 512);
-	assert(strlen(realpath) < 512);
+	assert(strlen(basepath_r) < 512);
 	if (basepath[0] != '/') {
 		fprintf(stderr, "testdir must be an absolute path\n");
 		return 1;
@@ -1880,10 +2004,10 @@ int main(int argc, char *argv[])
 	sprintf(subfile, "%s/subfile", testdir2);
 	sprintf(testsock, "%s/testsock", basepath);
 
-	sprintf(testfile_r, "%s/testfile", realpath);
-	sprintf(testfile2_r, "%s/testfile2", realpath);
-	sprintf(testdir_r, "%s/testdir", realpath);
-	sprintf(testdir2_r, "%s/testdir2", realpath);
+	sprintf(testfile_r, "%s/testfile", basepath_r);
+	sprintf(testfile2_r, "%s/testfile2", basepath_r);
+	sprintf(testdir_r, "%s/testdir", basepath_r);
+	sprintf(testdir2_r, "%s/testdir2", basepath_r);
 	sprintf(subfile_r, "%s/subfile", testdir2_r);
 
 	is_root = (geteuid() == 0);
@@ -1959,7 +2083,6 @@ int main(int argc, char *argv[])
 	err += test_create_ro_dir(O_CREAT | O_TRUNC);
 	err += test_copy_file_range();
 
-	unlink(testfile);
 	unlink(testfile2);
 	unlink(testsock);
 	rmdir(testdir);
@@ -1970,5 +2093,5 @@ int main(int argc, char *argv[])
 		return 1;
 	}
 
-	return 0;
+	return check_unlinked_testfiles();
 }
